#include "CelestialObject.h"

#include <GL/glut.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

#include "drawing.h"

const double G = 6.67430 * 1E-11;
const int32_t PREVIOUS_POSITIONS_MAXIMUM_LENGTH = 300;

CelestialObject *celestial_object_create(char *name, double mass, double semi_major_axis, double eccentricity, int32_t drawing_disc_radius, int32_t drawing_color) {
    CelestialObject *object = (CelestialObject *)malloc(sizeof(CelestialObject));

    strcpy(object->name, name);
    object->mass = mass;
    object->previous_position = vector2_create_zero();

    // Calculates the position at the periapsis.
    double periapsis = semi_major_axis * (1 - eccentricity);
    object->position = vector2_create(-periapsis, 0);

    object->semi_major_axis = semi_major_axis;
    object->eccentricity = eccentricity;

    object->drawing_disc_radius = drawing_disc_radius;
    object->drawing_color = drawing_color;

    object->previous_positions = (Vector2 *)malloc(sizeof(Vector2) * PREVIOUS_POSITIONS_MAXIMUM_LENGTH);
    object->previous_positions_length = 0;

    return object;
}

void celestial_object_destroy(CelestialObject *object) {
    free(object->previous_positions);
    free(object);
}

int32_t get_zoomed_drawing_disc_radius(CelestialObject *object, double zoom_factor) {
    if (zoom_factor < 1) {
        return object->drawing_disc_radius * zoom_factor;
    }

    return object->drawing_disc_radius;
}

Vector2 calculate_gravitational_acceleration(int32_t object_index, CelestialObject **objects, int32_t objects_length) {
    Vector2 a = vector2_create_zero();

    for (int32_t i = 0; i < objects_length; i += 1) {
        if (i == object_index) {
            // All except himself.
            continue;
        }

        Vector2 r = vector2_substract((i < object_index) ? objects[i]->previous_position : objects[i]->position, objects[object_index]->position);
        // Calculates the acceleration.
        double a_scalar = G * objects[i]->mass * pow(pow(vector2_norm(r), 2), -1);
        // Sum the acceleration.
        a = vector2_add(a, vector2_multiply(vector2_normalize(r), a_scalar));
    }

    return a;
}

void celestial_object_first_update(int32_t object_index, CelestialObject **objects, int32_t objects_length, int32_t main_object_index) {
    CelestialObject *object = objects[object_index];
    CelestialObject *main_object = objects[main_object_index];

    // Speed at the periapsis.
    double periapsis_speed = sqrt((G * main_object->mass * (1 + object->eccentricity)) / (object->semi_major_axis * (1 - object->eccentricity)));
    // Adds the speed of the main object.
    periapsis_speed += vector2_norm(vector2_substract(main_object->position, main_object->previous_position));

    Vector2 r = vector2_normalize(vector2_create(object->position.y, -object->position.x));
    Vector2 periapsis_velocity = vector2_multiply(r, periapsis_speed);

    // Adds the position of the main object.
    object->position = vector2_add(object->position, main_object->previous_position);

    // Calculates the new position of the object according to its velocity at periapsis.
    Vector2 new_position = vector2_add(object->position, periapsis_velocity);
    Vector2 a = calculate_gravitational_acceleration(object_index, objects, objects_length);
    // Apply the acceleration.
    new_position = vector2_add(new_position, vector2_multiply(a, 0.5));

    object->previous_position = object->position;
    object->position = new_position;
}

void celestial_object_update(int32_t object_index, CelestialObject **objects, int32_t objects_length, double interval, double previous_interval) {
    CelestialObject *object = objects[object_index];

    double interval_multiplier = interval / previous_interval;
    // Calculates its new position in relation to its previous and current position.
    Vector2 new_position = vector2_add(object->position, vector2_multiply(vector2_substract(object->position, object->previous_position), interval_multiplier));
    Vector2 a = calculate_gravitational_acceleration(object_index, objects, objects_length);
    // Adds the acceleration.
    new_position = vector2_add(new_position, vector2_multiply(a, pow(interval, 2)));

    object->previous_position = object->position;
    object->position = new_position;

    celestial_object_update_previous_positions(object);
}

void celestial_object_update_previous_positions(CelestialObject *object) {
    // Index 0 is always equal to the position of the object.
    object->previous_positions[0] = object->position;
    int32_t length = object->previous_positions_length;

    if (length > 1 && vector2_norm(vector2_substract(object->previous_positions[1], object->position)) < 1E9) {
        // If the distance between the position of the object and the point at index 1 is smaller than 1E9 then the position of the object is not added.
        return;
    }

    // Shifts all points to the right from index 1.
    for (int32_t j = (length == PREVIOUS_POSITIONS_MAXIMUM_LENGTH) ? length - 1 : length; j >= 2; j -= 1) {
        object->previous_positions[j] = object->previous_positions[j - 1];
    }

    // Sets the position of the object to index 1.
    object->previous_positions[1] = object->position;

    if (length < PREVIOUS_POSITIONS_MAXIMUM_LENGTH) {
        object->previous_positions_length += 1;
    }
}

void celestial_object_draw(CelestialObject *object, Vector2 reference_frame, double zoom_factor) {
    int32_t color = object->drawing_color;
    // Changes the color by decomposing the hexadecimal.
    glColor3ub((color & 0xFF0000) >> 16, (color & 0x00FF00) >> 8, (color & 0x0000FF) >> 0);

    Vector2 scaled_position = scale_position(object->position, reference_frame, zoom_factor);
    draw_disc(scaled_position, get_zoomed_drawing_disc_radius(object, zoom_factor));

    // Condition hard coded (should be dynamic to be cleaner).
    if ((strcmp(object->name, "Moon") != 0 && strcmp(object->name, "Apollo 11") != 0)) {
        draw_scaled_lines(object->previous_positions, object->previous_positions_length, reference_frame, zoom_factor);
    }
}

void celestial_object_draw_name(CelestialObject *object, Vector2 reference_frame, double zoom_factor) {
    Vector2 scaled_position = scale_position(object->position, reference_frame, zoom_factor);

    // Condition hard coded (should be dynamic to be cleaner).
    if (strcmp(object->name, "Moon") == 0 && zoom_factor < 36) {
        return;
    }

    // Condition hard coded (should be dynamic to be cleaner).
    if (strcmp(object->name, "Apollo 11") == 0 && zoom_factor < 2000) {
        return;
    }

    int32_t zoomed_drawing_disc_radius = get_zoomed_drawing_disc_radius(object, zoom_factor);
    // Draws the text just to the right of the object.
    draw_text(object->name, vector2_add(scaled_position, vector2_create(zoomed_drawing_disc_radius + 5, -10)));
}
