/**
 * @file CelestialObject.h
 * @author Gawen ACKERMANN and Florian BURGENER
 * @brief Set of functions and structures allowing the simulation of a celestial object.
 * @version 1.0
 * @date 2021-12-19
 *
 */

#ifndef CELESTIAL_OBJECT_H
#define CELESTIAL_OBJECT_H

#include "Vector2.h"

typedef struct CelestialObject {
    char name[100];
    double mass;
    Vector2 previous_position;
    Vector2 position;
    double semi_major_axis;
    double eccentricity;
    int32_t drawing_disc_radius;
    int32_t drawing_color;
    Vector2 *previous_positions;
    int32_t previous_positions_length;
} CelestialObject;

/**
 * @brief Creates a new CelestialObject.
 *
 * @param name
 * @param mass
 * @param semi_major_axis
 * @param eccentricity
 * @param drawing_disc_radius
 * @param drawing_color
 * @return CelestialObject*
 */
CelestialObject *celestial_object_create(char *name, double mass, double semi_major_axis, double eccentricity, int32_t drawing_disc_radius, int32_t drawing_color);

/**
 * @brief Destroys a PlanetarySystem.
 *
 * @param object
 */
void celestial_object_destroy(CelestialObject *object);

/**
 * @brief Gets the radius of the circle after applying the zoom.
 *
 * @param object
 * @param zoom_factor
 * @return int32_t
 */
int32_t get_zoomed_drawing_disc_radius(CelestialObject *object, double zoom_factor);

/**
 * @brief Calculates the gravitational acceleration applied to the object.
 *
 * @param object_index
 * @param objects
 * @param objects_length
 * @return Vector2
 */
Vector2 calculate_gravitational_acceleration(int32_t object_index, CelestialObject **objects, int32_t objects_length);

/**
 * @brief Calculates the first update.
 *
 * @param object_index
 * @param objects
 * @param objects_length
 * @param main_object_index
 */
void celestial_object_first_update(int32_t object_index, CelestialObject **objects, int32_t objects_length, int32_t main_object_index);

/**
 * @brief Updates the object.
 *
 * @param object_index
 * @param objects
 * @param objects_length
 * @param interval
 * @param previous_interval
 */
void celestial_object_update(int32_t object_index, CelestialObject **objects, int32_t objects_length, double interval, double previous_interval);

/**
 * @brief Updates the previous positions of the object.
 *
 * @param object
 */
void celestial_object_update_previous_positions(CelestialObject *object);

/**
 * @brief Draws the object.
 *
 * @param object
 * @param reference_frame
 * @param zoom_factor
 */
void celestial_object_draw(CelestialObject *object, Vector2 reference_frame, double zoom_factor);

/**
 * @brief Draws the name of the object.
 *
 * @param object
 * @param reference_frame
 * @param zoom_factor
 */
void celestial_object_draw_name(CelestialObject *object, Vector2 reference_frame, double zoom_factor);

#endif
