#include "PlanetarySystem.h"

#include <GL/glut.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "CelestialObject.h"
#include "Vector2.h"
#include "drawing.h"

const int32_t SCREEN_WIDTH = 800;
const int32_t SCREEN_HEIGHT = 800;

// Sun (https://nssdc.gsfc.nasa.gov/planetary/factsheet/sunfact.html)
const char SUN_NAME[] = "Sun";
const double SUN_MASS = 1988500 * 1E24;
const double SUN_SEMI_MAJOR_AXIS = 0;
const double SUN_ECCENTRICITY = 0;
const int32_t SUN_DRAWING_RADIUS = 50;
const int32_t SUN_COLOR = 0xFFCC33;
// Mercury (https://nssdc.gsfc.nasa.gov/planetary/factsheet/mercuryfact.html)
const char MERCURY_NAME[] = "Mercury";
const double MERCURY_MASS = 0.33010 * 1E24;
const double MERCURY_SEMI_MAJOR_AXIS = 57.909 * 1E9;
const double MERCURY_ECCENTRICITY = 0.2056;
const int32_t MERCURY_DRAWING_RADIUS = 10;
const int32_t MERCURY_COLOR = 0xDBCECA;
// Venus (https://nssdc.gsfc.nasa.gov/planetary/factsheet/venusfact.html)
const char VENUS_NAME[] = "Venus";
const double VENUS_MASS = 4.8673 * 1E24;
const double VENUS_SEMI_MAJOR_AXIS = 108.209 * 1E9;
const double VENUS_ECCENTRICITY = 0.0067;
const int32_t VENUS_DRAWING_RADIUS = 20;
const int32_t VENUS_COLOR = 0x8B7D82;
// Earth (https://nssdc.gsfc.nasa.gov/planetary/factsheet/earthfact.html)
const char EARTH_NAME[] = "Earth";
const double EARTH_MASS = 5.9722 * 1E24;
const double EARTH_SEMI_MAJOR_AXIS = 149.596 * 1E9;
const double EARTH_ECCENTRICITY = 0.0167;
const int32_t EARTH_DRAWING_RADIUS = 10;
const int32_t EARTH_COLOR = 0x6b93d6;
// Moon (https://nssdc.gsfc.nasa.gov/planetary/factsheet/moonfact.html)
const char MOON_NAME[] = "Moon";
const double MOON_MASS = 0.07346 * 1E24;
const double MOON_SEMI_MAJOR_AXIS = 0.3844 * 1E9;
const double MOON_ECCENTRICITY = 0.0549;
const int32_t MOON_DRAWING_RADIUS = 5;
const int32_t MOON_COLOR = 0x7D7B67;
// Apollo 11
const char APOLLO_11_NAME[] = "Apollo 11";
const double APOLLO_11_MASS = 28800;
const double APOLLO_11_SEMI_MAJOR_AXIS = 1E7 * 2;
const double APOLLO_11_ECCENTRICITY = 0;
const int32_t APOLLO_11_DRAWING_RADIUS = 2;
const int32_t APOLLO_11_COLOR = 0x0B3D91;
// Mars (https://nssdc.gsfc.nasa.gov/planetary/factsheet/marsfact.html)
const char MARS_NAME[] = "Mars";
const double MARS_MASS = 0.64169 * 1E24;
const double MARS_SEMI_MAJOR_AXIS = 227.923 * 1E9;
const double MARS_ECCENTRICITY = 0.0935;
const int32_t MARS_DRAWING_RADIUS = 12;
const int32_t MARS_COLOR = 0xBC2732;
// Jupiter (https://nssdc.gsfc.nasa.gov/planetary/factsheet/jupiterfact.html)
const char JUPITER_NAME[] = "Jupiter";
const double JUPITER_MASS = 1898.13 * 1E24;
const double JUPITER_SEMI_MAJOR_AXIS = 778.570 * 1E9;
const double JUPITER_ECCENTRICITY = 0.0489;
const int32_t JUPITER_DRAWING_RADIUS = 30;
const int32_t JUPITER_COLOR = 0xD8CA9D;
// Naboo
const char NABOO_NAME[] = "Naboo";
const double NABOO_MASS = 10 * 1E24;
const double NABOO_SEMI_MAJOR_AXIS = 100 * 1E9;
const double NABOO_ECCENTRICITY = 0.6;
const int32_t NABOO_DRAWING_RADIUS = 13;
const int32_t NABOO_COLOR = 0x00008B;
// Endor
const char ENDOR_NAME[] = "Endor";
const double ENDOR_MASS = 0.1 * 1E24;
const double ENDOR_SEMI_MAJOR_AXIS = 100 * 1E9;
const double ENDOR_ECCENTRICITY = 0.34;
const int32_t ENDOR_DRAWING_RADIUS = 8;
const int32_t ENDOR_COLOR = 0x228B22;
// Coruscant
const char CORUSCANT_NAME[] = "Coruscant";
const double CORUSCANT_MASS = 1E27;
const double CORUSCANT_SEMI_MAJOR_AXIS = 600 * 1E9;
const double CORUSCANT_ECCENTRICITY = 0.4675;
const int32_t CORUSCANT_DRAWING_RADIUS = 35;
const int32_t CORUSCANT_COLOR = 0x5C4033;

const int32_t SUN_INDEX = 0;
const int32_t EARTH_INDEX = 3;
const int32_t MOON_INDEX = 4;
const int32_t OBJECTS_LENGTH = 11;

PlanetarySystem *planetary_system_create() {
    PlanetarySystem *planetary_system = (PlanetarySystem *)malloc(sizeof(PlanetarySystem));

    planetary_system->objects = (CelestialObject **)malloc(sizeof(PlanetarySystem *) * OBJECTS_LENGTH);
    planetary_system->objects_length = OBJECTS_LENGTH;
    planetary_system->reference_frame_index = 0;
    planetary_system->zoom_factor = 1;
    planetary_system->previous_interval = 1;
    planetary_system->show_names = true;

    char *names[] = {(char *)SUN_NAME, (char *)MERCURY_NAME, (char *)VENUS_NAME, (char *)EARTH_NAME, (char *)MOON_NAME, (char *)APOLLO_11_NAME, (char *)MARS_NAME, (char *)JUPITER_NAME, (char *)NABOO_NAME, (char *)ENDOR_NAME, (char *)CORUSCANT_NAME};
    double masses[] = {SUN_MASS, MERCURY_MASS, VENUS_MASS, EARTH_MASS, MOON_MASS, APOLLO_11_MASS, MARS_MASS, JUPITER_MASS, NABOO_MASS, ENDOR_MASS, CORUSCANT_MASS};
    double semi_major_axes[] = {SUN_SEMI_MAJOR_AXIS, MERCURY_SEMI_MAJOR_AXIS, VENUS_SEMI_MAJOR_AXIS, EARTH_SEMI_MAJOR_AXIS, MOON_SEMI_MAJOR_AXIS, APOLLO_11_SEMI_MAJOR_AXIS, MARS_SEMI_MAJOR_AXIS, JUPITER_SEMI_MAJOR_AXIS, NABOO_SEMI_MAJOR_AXIS, ENDOR_SEMI_MAJOR_AXIS, CORUSCANT_SEMI_MAJOR_AXIS};
    double eccentricities[] = {SUN_ECCENTRICITY, MERCURY_ECCENTRICITY, VENUS_ECCENTRICITY, EARTH_ECCENTRICITY, MOON_ECCENTRICITY, APOLLO_11_ECCENTRICITY, MARS_ECCENTRICITY, JUPITER_ECCENTRICITY, NABOO_ECCENTRICITY, ENDOR_ECCENTRICITY, CORUSCANT_ECCENTRICITY};
    int32_t drawing_disc_radiuses[] = {SUN_DRAWING_RADIUS, MERCURY_DRAWING_RADIUS, VENUS_DRAWING_RADIUS, EARTH_DRAWING_RADIUS, MOON_DRAWING_RADIUS, APOLLO_11_DRAWING_RADIUS, MARS_DRAWING_RADIUS, JUPITER_DRAWING_RADIUS, NABOO_DRAWING_RADIUS, ENDOR_DRAWING_RADIUS, CORUSCANT_DRAWING_RADIUS};
    int32_t drawing_colors[] = {SUN_COLOR, MERCURY_COLOR, VENUS_COLOR, EARTH_COLOR, MOON_COLOR, APOLLO_11_COLOR, MARS_COLOR, JUPITER_COLOR, NABOO_COLOR, ENDOR_COLOR, CORUSCANT_COLOR};

    for (int32_t i = 0; i < planetary_system->objects_length; i += 1) {
        CelestialObject *object = celestial_object_create(names[i], masses[i], semi_major_axes[i], eccentricities[i], drawing_disc_radiuses[i], drawing_colors[i]);
        planetary_system->objects[i] = object;
    }

    int32_t main_object_indexes[] = {SUN_INDEX, SUN_INDEX, SUN_INDEX, SUN_INDEX, EARTH_INDEX, MOON_INDEX, SUN_INDEX, SUN_INDEX, SUN_INDEX, SUN_INDEX, SUN_INDEX};

    // Calculates the first position for all objects.
    for (int32_t i = 0; i < planetary_system->objects_length; i += 1) {
        if (i != 0) {
            celestial_object_first_update(i, planetary_system->objects, planetary_system->objects_length, main_object_indexes[i]);
        }
    }

    return planetary_system;
}

void planetary_system_destroy(PlanetarySystem *planetary_system) {
    // Destroys all objects.
    for (int32_t i = 0; i < planetary_system->objects_length; i += 1) {
        celestial_object_destroy(planetary_system->objects[i]);
    }

    free(planetary_system->objects);
    free(planetary_system);
}

Vector2 planetary_system_get_reference_frame(PlanetarySystem *planetary_system) {
    CelestialObject *object_reference_frame = planetary_system->objects[planetary_system->reference_frame_index];
    return object_reference_frame->position;
}

void planetary_system_update(PlanetarySystem *planetary_system, double interval) {
    // Updates all objects.
    for (int32_t i = 1; i < planetary_system->objects_length; i += 1) {
        celestial_object_update(i, planetary_system->objects, planetary_system->objects_length, interval, planetary_system->previous_interval);
    }

    planetary_system->previous_interval = interval;
}

void planetary_system_draw(PlanetarySystem *planetary_system) {
    Vector2 reference_frame = planetary_system_get_reference_frame(planetary_system);

    // Draws all objects.
    for (int32_t i = 0; i < planetary_system->objects_length; i += 1) {
        CelestialObject *object = planetary_system->objects[i];
        celestial_object_draw(object, reference_frame, planetary_system->zoom_factor);
    }

    // Drawing the names of the objects afterwards so that they are not under the rest of the drawings.
    planetary_system_draw_object_names(planetary_system);

    char text[200];
    sprintf(text, "Focused Object : %s", planetary_system->objects[planetary_system->reference_frame_index]->name);
    draw_text(text, vector2_create(5, 15));
}

void planetary_system_draw_object_names(PlanetarySystem *planetary_system) {
    if (!planetary_system->show_names) {
        return;
    }

    Vector2 reference_frame = planetary_system_get_reference_frame(planetary_system);

    for (int32_t i = 0; i < planetary_system->objects_length; i += 1) {
        celestial_object_draw_name(planetary_system->objects[i], reference_frame, planetary_system->zoom_factor);
    }
}
