/**
 * @file main.c
 * @author Gawen ACKERMANN and Florian BURGENER
 * @brief Simulation of a planetary system.
 * @version 1.0
 * @date 2021-12-19
 *
 */

#include <GL/freeglut.h>
#include <GL/glut.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>

#include "PlanetarySystem.h"
#include "drawing.h"

// Name of the window.
const char WINDOW_NAME[] = "Planetary System";
// Refresh rate in Hz.
const int32_t REFRESH_RATE = 200;
// Number of microseconds in 1 second.
const int32_t ONE_SECOND_IN_MICROSECONDS = 1E6;
// Number of milliseconds in 1 second.
const int32_t ONE_SECOND_IN_MILLISECONDS = 1000;
// Number of seconds in a day.
const int32_t ONE_DAY_IN_SECONDS = 86400;
// Zoom multiplier at each wheel movement.
const double ZOOM_MULTIPLIER = 1.1;
// Default simulation speed in days.
const int32_t DEFAULT_SIMULATION_SPEED_IN_DAYS = 10;
// Maximum speed of the simulation in days.
const int32_t SIMULATION_MAXIMUM_SPEED_IN_DAYS = 500;
// Time that elapses by default in the planetary system per second, in seconds.
const int32_t DEFAULT_SIMULATION_SPEED_IN_SECONDS = DEFAULT_SIMULATION_SPEED_IN_DAYS * ONE_DAY_IN_SECONDS;
// Maximum time that elapses in the planetary system at each call of the planetary_system_update function, this value must be small enough for the planetary system to work.
const int32_t PLANETARY_SYSTEM_INTERVAL = 1000;

#define GLUT_KEY_ESCAPE 27
#define GLUT_KEY_T 116

// https://stackoverflow.com/questions/3417837/what-is-the-best-way-to-suppress-a-unused-variable-x-warning
#ifdef UNUSED
#elif defined(__GNUC__)
#define UNUSED(x) UNUSED_##x __attribute__((unused))
#elif defined(__LCLINT__)
#define UNUSED(x) /*@unused@*/ x
#else
#define UNUSED(x) x
#endif

PlanetarySystem *planetary_system;
int32_t simulation_speed = DEFAULT_SIMULATION_SPEED_IN_SECONDS;
int64_t start_time;
int64_t previous_time;

/**
 * @brief Get the current time in microseconds.
 *
 * @return int64_t
 */
int64_t get_current_time() {
    struct timeval time;
    gettimeofday(&time, NULL);
    return (int64_t)time.tv_sec * (int64_t)ONE_SECOND_IN_MICROSECONDS + (int64_t)time.tv_usec;
}

/**
 * @brief Updates the planetary system.
 *
 */
void update() {
    int64_t current_time = get_current_time();
    int64_t elapsed_time = current_time - previous_time;
    // Time that must elapse in the simulation at this frame.
    double remaining_time = (double)elapsed_time / ONE_SECOND_IN_MICROSECONDS * simulation_speed;

    while (remaining_time >= PLANETARY_SYSTEM_INTERVAL) {
        // Move the planetary system forward by PLANETARY_SYSTEM_INTERVAL seconds while there is still time to go.
        planetary_system_update(planetary_system, PLANETARY_SYSTEM_INTERVAL);
        remaining_time -= PLANETARY_SYSTEM_INTERVAL;
    }

    if (remaining_time > 0) {
        planetary_system_update(planetary_system, remaining_time);
    }

    previous_time = current_time;
}

/**
 * @brief Timer for the "update" function.
 *
 */
void update_timer() {
    update();
    glutTimerFunc(ONE_SECOND_IN_MILLISECONDS / REFRESH_RATE, update_timer, 0);
}

/**
 * @brief Draws the planetary system.
 *
 */
void draw() {
    glClearColor(0.0, 0.0, 0.0, 1.0);
    glClear(GL_COLOR_BUFFER_BIT);

    char title[100];
    double elapsed_time = (double)(get_current_time() - start_time) / ONE_SECOND_IN_MICROSECONDS;
    sprintf(title, "%s (%.2lf)", WINDOW_NAME, elapsed_time);
    glutSetWindowTitle(title);

    planetary_system_draw(planetary_system);

    char text[100];
    sprintf(text, "Simulation Speed : %d days per second", simulation_speed / ONE_DAY_IN_SECONDS);
    draw_text(text, vector2_create(5, 0));

    glutSwapBuffers();
}

/**
 * @brief Timer for the "draw" function.
 *
 */
void draw_timer() {
    glutPostRedisplay();
    glutTimerFunc(ONE_SECOND_IN_MILLISECONDS / REFRESH_RATE, draw_timer, 0);
}

/**
 * @brief Event that retrieves the keyboard inputs.
 *
 * @param key
 * @param UNUSED
 * @param UNUSED
 */
void handle_keyboard_input(unsigned char key, int UNUSED(x), int UNUSED(y)) {
    if (key == GLUT_KEY_ESCAPE) {
        // Quits the program.
        glutLeaveMainLoop();
        return;
    }

    if (key == GLUT_KEY_T) {
        // Toggles the display of object names.
        planetary_system->show_names = !planetary_system->show_names;
    }
}

/**
 * @brief Event that retrieves the special keyboard inputs.
 *
 * @param key
 * @param UNUSED
 * @param UNUSED
 */
void handle_special_keyboard_input(int key, int UNUSED(x), int UNUSED(y)) {
    if (key == GLUT_KEY_LEFT) {
        // Follows the previous object.
        planetary_system->reference_frame_index -= 1;

        if (planetary_system->reference_frame_index < 0) {
            planetary_system->reference_frame_index = planetary_system->objects_length - 1;
        }
    }

    if (key == GLUT_KEY_RIGHT) {
        // Follows the next object.
        planetary_system->reference_frame_index += 1;
        planetary_system->reference_frame_index %= planetary_system->objects_length;
    }

    if (key == GLUT_KEY_UP) {
        // Increases the simulation speed.
        simulation_speed += ONE_DAY_IN_SECONDS;

        if (simulation_speed > ONE_DAY_IN_SECONDS * SIMULATION_MAXIMUM_SPEED_IN_DAYS) {
            simulation_speed = ONE_DAY_IN_SECONDS * SIMULATION_MAXIMUM_SPEED_IN_DAYS;
        }
    }

    if (key == GLUT_KEY_DOWN) {
        // Decreases the simulation speed.
        simulation_speed -= ONE_DAY_IN_SECONDS;

        if (simulation_speed < 0) {
            simulation_speed = 0;
        }
    }
}

/**
 * @brief Event that retrieves actions performed with the mouse.
 *
 * @param button
 * @param state
 * @param UNUSED
 * @param UNUSED
 */
void handle_mouse_input(int button, int state, int UNUSED(x), int UNUSED(y)) {
    if (button == 3 || button == 4) {
        if (state == GLUT_UP) return;

        if (button == 3) {
            // Zoom in.
            planetary_system->zoom_factor *= ZOOM_MULTIPLIER;
        } else {
            // Zoom out.
            planetary_system->zoom_factor /= ZOOM_MULTIPLIER;
        }
    }
}

/**
 * @brief Entry point of the program.
 *
 * @param argc
 * @param argv
 * @return int
 */
int main(int argc, char *argv[]) {
    planetary_system = planetary_system_create();

    glutInit(&argc, argv);
    glutSetOption(GLUT_MULTISAMPLE, 16);
    glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_MULTISAMPLE);
    glutInitWindowSize(SCREEN_WIDTH, SCREEN_HEIGHT);
    int window = glutCreateWindow(WINDOW_NAME);

    glClearColor(0.0, 0.0, 0.0, 1.0);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(0, SCREEN_WIDTH, SCREEN_HEIGHT, 0);

    glutDisplayFunc(draw);
    glutKeyboardFunc(handle_keyboard_input);
    glutSpecialFunc(handle_special_keyboard_input);
    glutMouseFunc(handle_mouse_input);

    start_time = get_current_time();
    previous_time = start_time;

    glutTimerFunc(ONE_SECOND_IN_MILLISECONDS / REFRESH_RATE, update_timer, 0);
    glutTimerFunc(ONE_SECOND_IN_MILLISECONDS / REFRESH_RATE, draw_timer, 0);

    glutMainLoop();

    // Destroys the window and the planetary system.
    glutDestroyWindow(window);
    planetary_system_destroy(planetary_system);

    return EXIT_SUCCESS;
}
